"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const sort_1 = require("./sort");
class System extends events_1.EventEmitter {
    constructor(options) {
        super();
        this.components = [];
        this.running = false;
        this.setOptions(options || {});
        this.listenSignals();
    }
    setOptions({ name, exitOnError = true, terminationTimeout = 3000 }) {
        this.options = { name, exitOnError, terminationTimeout };
    }
    listenSignals() {
        process.on('message', this.handleProcessMessage.bind(this));
        process.on('SIGINT', this.handleSignal.bind(this, 'SIGINT'));
        process.on('SIGTERM', this.handleSignal.bind(this, 'SIGTERM'));
        process.on('uncaughtException', this.handleUncaughtException.bind(this));
        process.on('unhandledRejection', this.handleUnhandledRejection.bind(this));
        process.on('exit', this.handleExit.bind(this));
    }
    updateLast(delta) {
        this.components = this.components.map((component, i) => i === this.components.length - 1 ? Object.assign({}, component, delta) : component);
    }
    dependsOn(...deps) {
        this.updateLast({ dependencies: deps.map(createDependency) });
        return this;
    }
    ignorable() {
        this.updateLast({ mandatory: false });
        return this;
    }
    add(name, component) {
        this.components = [
            ...this.components,
            Object.assign({}, component, { name, dependencies: [], mandatory: true })
        ];
        return this;
    }
    start(initResources = {}, restart = this.restart.bind(this), interrupt = this.handleInterruption.bind(this)) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const resources = yield start(sort_1.default(this.components), initResources, restart, interrupt, (componentName, resources) => this.emit('componentStart', componentName, resources), (componentName, error) => this.emit('componentStartFailed', componentName, error));
                this.running = true;
                this.emit('start', resources);
                return resources;
            }
            catch (error) {
                yield this.gracefulTerminate(error);
                return {};
            }
        });
    }
    stop() {
        return __awaiter(this, void 0, void 0, function* () {
            yield stop(sort_1.default(this.components).reverse(), (componentName) => this.emit('componentStop', componentName), (componentName, error) => this.emit('componentStopFailed', componentName, error));
            this.running = false;
            this.emit('stop');
        });
    }
    restart(component) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.stop();
            const resources = yield this.start();
            this.emit('restart', resources);
            return resources;
        });
    }
    group() {
        return {
            start: (initResources, restart, stop) => __awaiter(this, void 0, void 0, function* () { return filterResources(initResources, yield this.start(initResources, restart, stop)); }),
            stop: () => __awaiter(this, void 0, void 0, function* () { return yield this.stop(); })
        };
    }
    logAllEvents() {
        const systemName = this.options.name;
        return this
            .on('start', resources => console.log(`${systemName || 'System'} started`))
            .on('stop', () => console.log(`${systemName || 'System'} stopped`))
            .on('restart', resources => console.log(`${systemName || 'System'} restarted`))
            .on('componentStart', (name, resources) => console.log(`Component started: ${name}`))
            .on('componentStartFailed', (name, err) => console.log(`Component start failed: ${name}`, err))
            .on('componentStop', name => console.log(`Component stopped: ${name}`))
            .on('componentStopFailed', (name, err) => console.log(`Component stop failed: ${name}`, err))
            .on('terminationTimeout', timeout => console.log(`Stop timeout: ${timeout}`))
            .on('componentRunFailed', (name, err) => console.log(`Component run failed: ${name}`, err))
            .on('uncaughtException', err => console.log(`UncaughtException`, err))
            .on('unhandledRejection', err => console.log(`UnhandledRejection`, err))
            .on('signal', signal => console.log(signal))
            .on('exit', code => console.log(`Exit with code: ${code}`));
    }
    handleInterruption(component, error) {
        return __awaiter(this, void 0, void 0, function* () {
            this.emit('componentRunFailed', component.name, error);
            if (component.mandatory) {
                this.gracefulTerminate(error);
            }
        });
    }
    handleProcessMessage(msg) {
        if (msg === 'shutdown') {
            this.emit('signal', msg);
            this.gracefulTerminate();
        }
    }
    handleSignal(signal) {
        this.emit('signal', signal);
        this.gracefulTerminate();
    }
    handleUncaughtException(error) {
        this.emit('uncaughtException', error);
        process.once('uncaughtException', this.handleError.bind(this));
        this.gracefulTerminate(error);
    }
    handleUnhandledRejection(error) {
        this.emit('unhandledRejection', error);
        process.once('unhandledRejection', this.handleError.bind(this));
        this.gracefulTerminate(error);
    }
    handleTerminationTimeout() {
        this.emit('terminationTimeout', this.options.terminationTimeout);
        this.handleError(new Error("Termination Timeout"));
    }
    handleError(error) {
        if (error && this.options.exitOnError)
            this.exit(1);
    }
    handleExit(code) {
        this.emit('exit', code);
    }
    gracefulTerminate(error) {
        return __awaiter(this, void 0, void 0, function* () {
            setTimeout(this.handleTerminationTimeout.bind(this), this.options.terminationTimeout);
            if (this.running)
                yield this.stop();
            error ? this.handleError(error) : this.exit(0);
        });
    }
    exit(code) {
        process.exit(code);
    }
}
exports.System = System;
exports.default = System;
function createDependency(dep) {
    switch (typeof dep) {
        case 'string': return { component: dep, as: dep };
        case 'object':
            const { component, as, source } = dep;
            if (!component)
                throw new Error(`'component' is required property on dependency component`);
            return { component, as: as || component, source };
        default: throw new Error(`Invalid dependency component: ${dep}`);
    }
}
exports.createDependency = createDependency;
function mapResources(allResources, dependencies) {
    const resources = {};
    Object.keys(allResources).forEach(resourceName => {
        const ownDependency = dependencies.find(dep => dep.component === resourceName);
        if (ownDependency) {
            const { component, as, source } = ownDependency;
            resources[as] = source ? allResources[component][source] : allResources[component];
        }
        else {
            resources[resourceName] = allResources[resourceName];
        }
    });
    return resources;
}
exports.mapResources = mapResources;
function filterResources(initResources, subSystemResources) {
    const resources = Object.assign({}, subSystemResources);
    Object.keys(initResources).forEach(resourceName => delete resources[resourceName]);
    return resources;
}
exports.filterResources = filterResources;
function start([first, ...others], resources, restart, stop, onComponentStart, onComponentStartFailed) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!first)
            return resources;
        let resource;
        try {
            resource = yield first.start(mapResources(resources, first.dependencies), restart.bind(null, first), stop.bind(null, first));
        }
        catch (err) {
            onComponentStartFailed(first.name, err);
            if (first.mandatory)
                throw err;
            return yield start(others, resources, restart, stop, onComponentStart, onComponentStartFailed);
        }
        const extendedResources = Object.assign({}, resources, { [first.name]: resource });
        onComponentStart(first.name, extendedResources);
        return yield start(others, extendedResources, restart, stop, onComponentStart, onComponentStartFailed);
    });
}
exports.start = start;
function stop([first, ...others], onComponentStop, onComponentStopFailed) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!first)
            return;
        try {
            if (first.stop)
                yield first.stop();
            onComponentStop(first.name);
        }
        catch (err) {
            onComponentStopFailed(first.name, err);
        }
        return yield stop(others, onComponentStop, onComponentStopFailed);
    });
}
exports.stop = stop;
